#! /usr/bin/env python
#
#    Simple easily extended user interface for viewing models
#    Copyright (C) 2002  Michael Urman
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""Handle user interface for displaying models in an easily extended manner"""

import sys, math
try:
    import pygame
    from pygame.locals import *
    from OpenGL.GL import *
except:
    print 'model requires pygame and pyopengl'
    raise SystemExit

try:
    from OpenGL.GLU import *
    GLU = 1
except:
    print "Warning: OpenGL.GLU did not import correctly."
    GLU = None

import sprite

class Handler:
    """Define event handlers by matching all mask's key:val => call callback"""

    def __init__(self, mask, callback):
        """Handler(self, mask, callback) -> Handler

        Create a Handler with mask and callback.  mask should be a dict, and
        callback a function with an event argument"""

        self.mask = mask
        self.callback = callback

class UI:
    """Base User Interface"""
    pass

class pygameUI(UI):
    """PYGAME User Interface"""

    def add_model(self, model, name=None):
        """add_model(self, model, name=None) -> None

        Add a model to the scene.  A name is required for all models or no
        models if more than one is loaded.  Raises KeyError on repeated name.
        Raises TypeError if a name/no name conflict occurs."""

        if name == None:
            if self.models == None:
                self.active_model = model
                self.models = [model]
            else:
                try:
                    self.models.append(model)
                except AttributeError:
                    raise TypeError("Model requires a name")
        else:
            if self.models == None:
                self.active_model = model
                self.models = { name : model }
            else:
                try:
                    if self.models.has_key(name):
                        raise KeyError("Model %s already in use" % name)
                    else:
                        self.models[name] = model
                except AttributeError:
                    raise TypeError("Model must not have a name")

        # improve rendering efficiency umpteenfold
        # (python is a lot of overhead!)
        # model.build_display_list() # XXX ?

    def del_model(self, name=-1):
        """del_model(self, name=None) -> None

        Delete a model from the scene.  A name is required if models are named.
        An index is optional if models are numbered; otherwise models are
        removed in the reverse order they were added.  Raises KeyError on
        nonexistent name, IndexError on invalid index or empty modelset, or
        TypeError on name/no name mismatch.  Once all models are removed,
        add_model can switch to/from taking names."""

        if type(name) == type(''):
            try:
                if self.models[name] == self.active_model:
                    self.active_model = None
                del self.models[name]
            except KeyError:
                raise KeyError("Model %s not available" % name)
            except TypeError:
                raise TypeError("Cannot remove model without an index")
        else:
            if name < 0:
                name = len(self.models)-name
            try:
                if self.models[name] == self.active_model:
                    self.active_model = None
                self.models.pop(name)
                if len(self.models) == 0:
                    self.models = None
            except AttributeError:
                raise KeyError("Cannot remove model without a name")
            except IndexError:
                raise IndexError("Cannot remove model %s" % name)
            except TypeError:
                raise IndexError("No models to remove")

    def run(self, **kvargs):
        """run(self, **kvargs) -> None

        Initializes the display and runs until finished.
        
        Optional keywords include size:(x,y), and fullscreen:bool"""

        winsize = kvargs.has_key('size') and kvargs['size'] or (640, 480)
        fullscreen = kvargs.has_key('fullscreen') and kvargs['fullscreen'] and FULLSCREEN

        screen = pygame.display.set_mode(winsize, fullscreen|OPENGL|DOUBLEBUF)

        self.setup_gl()

        while 1:
            events = [pygame.event.wait()]
            if pygame.event.peek():
                events.extend(pygame.event.get())

            # process all events
            for event in events:
                for handler in self.handlers:
                    try:
                        for k,v in handler.mask.items():
                            if getattr(event, k) != v:
                                break
                        else:
                            if handler.callback(event):
                                break
                    except AttributeError:
                        pass

            # quit?
            if self.should_quit:
                break

            # draw?
            if self.should_update:
                self.should_update = 0
                self._cb_apply_mouse_motion()   # catch up to mouse

                glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
                try:
                    for model in self.models:
                        model.draw()
                except TypeError:
                    for name, model in self.models.items():
                        model.draw()
                #model.draw(wireframe=self.should_wireframe)

                if self.caption_show:
                    cs = self.caption_sprite
                    cs.enter()
                    cs.blit_at(
                        (screen.get_width()-cs.get_width()+1,
                        screen.get_height()-cs.get_height()+1))
                    cs.exit()

                pygame.display.flip()


    def cb_update(self, event):
        """set update flag"""
        self.should_update = 1

    def cb_quit(self, event):
        """set to quit"""
        self.should_quit = 1

    def cb_toggle_fullscreen(self, event):
        """toggle fullscreen (only on ALT-enter)"""
        if pygame.key.get_mods() & KMOD_ALT:
            pygame.display.toggle_fullscreen()
            self.cb_update(None)

    def cb_toggle_wireframe(self, event):
        """toggle active model's draw_wires attribute"""
        self.active_model.draw_wires = not self.active_model.draw_wires
        self.cb_update(None)

    def cb_toggle_faces(self, event):
        """toggle active model's draw_faces attribute"""
        self.active_model.draw_faces = not self.active_model.draw_faces
        self.cb_update(None)

    def cb_toggle_duals(self, event):
        """toggle active model's draw_duals attribute"""
        self.active_model.draw_duals = not self.active_model.draw_duals
        self.cb_update(None)

    def cb_toggle_normals(self, event):
        """toggle active model's draw_normals attribute"""
        self.active_model.draw_normals = not self.active_model.draw_normals
        self.cb_update(None)

    def cb_toggle_textures(self, event):
        """toggle active model's draw_faces attribute"""
        self.active_model.draw_textures = not self.active_model.draw_textures
        self.cb_update(None)

    def cb_defaultmouse(self, event):
        """default mouse handling"""
        if event.type == MOUSEBUTTONDOWN:
            if event.button in (1,2,3):
                self._cb_apply_mouse_motion()
                if not self.mouse_hidden:
                    pygame.mouse.set_visible(0)
                    pygame.event.set_grab(1)
                self.mouse_hidden += 1
        elif event.type == MOUSEBUTTONUP:
            if event.button in (1,2,3):
                self._cb_apply_mouse_motion()
                self.mouse_hidden -= 1
                if not self.mouse_hidden:
                    pygame.mouse.set_visible(1)
                    pygame.event.set_grab(0)
        elif event.type == MOUSEMOTION:
            if self.mouse_hidden:
                self.motion_buttons = event.buttons
                self.motion[0] += event.rel[0]
                self.motion[1] += event.rel[1]
                self.cb_update(None)

    def cb_select_next_model(self, event):
        delta = 1
        if pygame.key.get_mods() & KMOD_SHIFT:
            delta = -1
        models = self.models
        try:
            models = models.values()
        except AttributeError:
            pass

        am = self.active_model
        self.active_model = models[(models.index(am)+delta) % len(models)]


    def _cb_apply_mouse_motion(self):
        dx, dy = self.motion
        self.motion = [0,0]

        if self.motion_buttons[0]: # rotate
            dist = math.sqrt(dx*dx+dy*dy)
            self.rotate_active_model(dist, (dy, dx, 0))
        elif self.motion_buttons[1]: # scale
            s = math.exp((dx+dy)*0.01)
            self.scale_active_model((s,s,s))
        elif self.motion_buttons[2]: # translate
            self.translate_active_model((dx*0.02, -dy*0.02, 0))

    def scale_active_model(self, S):
        """scale_active_model(self, (x,y,z)) -> None

        Scale the active model by 3-tuple (x,y,z)"""

        self.active_model.scale(S)

    def rotate_active_model(self, theta, axis):
        """rotate_active_model(self, theta, (x,y,z)) -> None

        Rotate the active model by theta around the 3-tuple (x,y,z)"""

        self.active_model.rotate(theta, axis)

    def translate_active_model(self, T):
        """translate_active_model(self, (x,y,z)) -> None

        Translate the active model by 3-tuple (x,y,z)"""

        self.active_model.translate(T)

    def setup_gl(self):
        """Set up OpenGL to a simple, sane, state"""

        glEnable(GL_DEPTH_TEST)     # use zbuffer
        glEnable(GL_NORMALIZE)      # normalize normals, etc.

        # turn on ambient light
        glEnable(GL_LIGHTING)
        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, (1.0, 1.0, 1.0, 1.0))
        glLightfv(GL_LIGHT0, GL_POSITION, (0.0, 0.5, 1.0, 0.0))
        glEnable(GL_LIGHT0)
        glLightfv(GL_LIGHT1, GL_POSITION, (-0.8, -0.3, -0.7, 0.0))
        glEnable(GL_LIGHT1)

        # set material to a goldish tone
        rgb = (0.912, 0.717, 0.505, 1.0)
        r_amb = tuple(map(lambda x: 0.1*x, rgb))
        r_diff = tuple(map(lambda x: 1.0*x, rgb))
        r_spec = tuple(map(lambda x: 0.3*x, rgb))

        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, r_amb)
        glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, r_diff)
        glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, r_spec)
        glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, 100)

        # set up a simple view: pulled back and rotated up a bit
        glMatrixMode(GL_PROJECTION)
        if GLU:
            gluPerspective(45.0, 640/480.0, 0.1, 100.0)
        else:
            f = 1.3 / (math.tan(45.0/2))
            glMultMatrix((f*480/640, 0, 0, 0,
                        0, f, 0, 0,
                        0, 0, 100.1/-99.9, -1,
                        0, 0, 2*100*0.1/-99.9, 0))
        
        glTranslatef(0.0, 0.0, -3.0)
        glRotatef(25, 1,0,0)

    def set_caption(self, caption, color=(1,1,1), font=None, size=48):
        """set_caption(self, caption, color=(1,1,1), font=None, size=48) -> None
        
        Set window caption to caption, with color (default=white)"""

        self.caption_text = caption
        self.caption_color = color
        usefont = pygame.font.Font(font, size)
        self.caption_image = usefont.render(caption, 1, color)
        self.caption_sprite = sprite.sprite(self.caption_image)

    def show_caption(self, show):
        """show_caption(self, show) -> None

        Set caption to visible or invisible"""

        self.caption_show = not not show

    def add_handler(self, handler):
        self.handlers.append(handler)

    def clear_handlers(self):
        self.handlers = []

    def __init__(self):
        """__init__(self) -> pygameUI

        Initialize a pygameUI with default settings, handlers"""

        pygame.display.init()
        pygame.font.init()

        # default values
        self.set_caption('model.UI')
        self.caption_show = 0
        self.models = None
        self.should_quit = 0
        self.should_update = 1
        self.mouse_hidden = 0
        self.motion = [0,0]
        self.motion_buttons=[0,0,0]
        self.active_model = None
        # default handlers
        self.default_handlers = [
            # quit on close or ESC
            Handler({'type':QUIT}, self.cb_quit),
            Handler({'type':KEYDOWN, 'key':K_ESCAPE}, self.cb_quit),

            # toggle fullscreen on RETURN
            Handler({'type':KEYDOWN, 'key':K_RETURN}, self.cb_toggle_fullscreen),

            # toggle wireframe on W, faces on F, textures on T
            Handler({'type':KEYDOWN, 'key':K_w}, self.cb_toggle_wireframe),
            Handler({'type':KEYDOWN, 'key':K_f}, self.cb_toggle_faces),
            Handler({'type':KEYDOWN, 'key':K_t}, self.cb_toggle_textures),
            Handler({'type':KEYDOWN, 'key':K_d}, self.cb_toggle_duals),
            Handler({'type':KEYDOWN, 'key':K_n}, self.cb_toggle_normals),

            # next model selected on TAB
            Handler({'type':KEYDOWN, 'key':K_TAB}, self.cb_select_next_model),

            # redraw when exposed
            Handler({'type':VIDEOEXPOSE}, self.cb_update),

            # mouse handlers
            Handler({'type':MOUSEBUTTONDOWN}, self.cb_defaultmouse),
            Handler({'type':MOUSEBUTTONUP}, self.cb_defaultmouse),
            Handler({'type':MOUSEMOTION}, self.cb_defaultmouse),
        ]

        self.handlers = self.default_handlers[:]
